
#include <OpenCV/OpenCV.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>       // for memset
#include <unistd.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <BlobResult.h>

#include <fstream>
#include <iostream>
using namespace std;

#include "mainHelper.h"    /* defines ColorDefinition */

/*
 * global data for passing among 
 *   the main loop (in main)
 *   the mouse-event handler (in mouseEventHandler)
 *   the server (in runServerThread)
 */
IplImage *inputImage=0;  // shared between mouse events and main loop
int swapVerticalValues = 0;  // shared between mouse events and main loop

int drawingRectangle=0;  // shared between mouse events and main loop
CvPoint origin;          // more rectangle-selection data
CvRect selection;

int runAlgorithm = 0;    // shared between server and main loop
int timeToQuitNow = 0;   // shared between server and main loop

ColorDefinition currentColor;  // shared among mouse events, server, and main loop


#include "serverStuff.h"


/*
 * Mouse-event handling routine
 */
void mouseEventHandler( int event, int x, int y, int flags, void* param )
{
  int r,g,b;
  double h,s,v;
	
	if ( !inputImage )  return;  // no image, do nothing
  if (swapVerticalValues)  y = inputImage->height - y;   // platform-dependent!
  
  if ( event == CV_EVENT_MOUSEMOVE && drawingRectangle ) // are we pulling the corner of a rectangle?
  {
    // handle the geometry of the rectangle:
    selection.x = MIN(x,origin.x);  selection.y = MIN(y,origin.y);
    selection.width = MIN( selection.x + CV_IABS(x - origin.x), inputImage->width ) - MAX( selection.x, 0 );
    selection.height = MIN( selection.y + CV_IABS(y - origin.y), inputImage->height ) - MAX( selection.y, 0 );
  } // end of rectangle-drawing
    
  if ( event == CV_EVENT_RBUTTONDOWN ) // did we right-click (or control-click?)
  {
    getrgb( inputImage, x, y, r, g, b );
    RGBtoHSV( r, g, b, h, s, v );
    printf("      right-click at (%d,%d): rgb=(%d,%d,%d)\n", x, y, r, g, b ) ;
    printf("         hsv=(%6.3f,%6.3f,%6.3f)\n", h, s, v );
    int N = 10;
    currentColor.reset();
    currentColor.expandToFit( r+N, g+N, b+N ); // box of side 2N: upper corner
    currentColor.expandToFit( r-N, g-N, b-N ); // and lower corner
    currentColor.print();
  }
      
  if ( event == CV_EVENT_LBUTTONDOWN ) // did we left-click?
  {
    if (!drawingRectangle) // if we weren't drawing, now we start drawing a rectangle
    {
      origin = cvPoint(x,y);
      selection = cvRect(x,y,1,1);
      drawingRectangle = 1;
    }
    else // if we were drawing a rectangle, now we're finished drawing it
    {
      currentColor.reset();
      for ( int col=selection.x ; col<=selection.x+selection.width ; ++col)
      {
        for ( int row=selection.y ; row<=selection.y+selection.height ; ++row) 
        {
          getrgb( inputImage, col, row, r, g, b );
          currentColor.expandToFit( r, g, b );
        }
      }
      currentColor.print();
      drawingRectangle = 0;
    }
  }
} // end of mouse event handler


/*
 * MAIN
 */
int main( int argc, char** argv )
{	
  /*
   * set up vision server
   */
  pthread_t serverThread;   // the identifier for the thread
  int runServer = true;     
  
  if (runServer)            // do we want to run the server?
  {
    printf("Creating serverThread.\n");
    int rc = pthread_create(&serverThread, NULL, runServerThread, NULL);
    if (rc) printf("ERROR in pthread_create: %d.\n", rc);
  }
  
  /*
   * set up image input source
   */
  enum { IMAGES_FROM_CAMERA, IMAGES_FROM_FOLDER, IMAGES_FROM_AVI_FILE };
  int captureType = IMAGES_FROM_FOLDER;
  
  char* MOVIE_NAME = "../../../../gatorJeepMap.mpeg"; //  "../../../../jeep_onboard.avi"
  char* FOLDER_NAME = "../../../../VideoBMPs/";
  int MAX_FILE_NUMBER = 81;  // the number of images in the folder
  int nextFileNumber = 0;    // the next file number to read 
  
  enum { LIVE, ONEFRAME, STILL };  // how are we now capturing
  int captureStatus = ONEFRAME;
  
  CvCapture* capture = 0;  // the capture object to grab frames from a camera/avi file

  if ( captureType == IMAGES_FROM_CAMERA )   // initialize "capture," the image source
    capture = cvCaptureFromCAM( 0 );         // camera number zero, which is usually the one attached
  else if ( captureType == IMAGES_FROM_AVI_FILE )
    capture = cvCaptureFromAVI( MOVIE_NAME ); 
  else // captureType = IMAGES_FROM_FOLDER;
    capture = 0;
  
  /*
   * create windows for displaying images
   */
  cvNamedWindow( "OutputImage", CV_WINDOW_AUTOSIZE );
  cvNamedWindow( "InputImage", CV_WINDOW_AUTOSIZE );
	cvSetMouseCallback( "InputImage", mouseEventHandler, 0 );
  cvMoveWindow( "OutputImage", 100, 100 );
  cvMoveWindow( "InputImage", 150, 150 );
  
  cvNamedWindow( "red", CV_WINDOW_AUTOSIZE );
  cvMoveWindow( "red", 200,200 );
  cvNamedWindow( "gre", CV_WINDOW_AUTOSIZE );
  cvMoveWindow( "gre", 200,200 );
  cvNamedWindow( "blu", CV_WINDOW_AUTOSIZE );
  cvMoveWindow( "blu", 200,200 );
  
  /*
   * create an initial image, displayImage, and any other images...
   */
  inputImage = cvCreateImage( cvSize(10,10), 8, 1 );
  IplImage* outputImage = NULL;
  IplImage* frame = NULL;
  IplImage* blu = NULL;
  IplImage* gre = NULL;
  IplImage* red = NULL;
  
  CBlobResult blobs; // OK
  CBlob blobToShow;
  int nthblob = 0;
  int imageAreaInPixels = 100;
  CvFont ourFont;
  cvInitFont( &ourFont, CV_FONT_HERSHEY_SIMPLEX, 0.5, 0.5, 0.0, 2, CV_AA );
  
  CvMat* map_matrix;
  map_matrix = cvCreateMat( 3, 3, CV_64FC1 ); // CV_<bit_depth>(S|U|F)C<number_of_channels>
  double th = 45.0   * M_PI/180.0;
  cvmSet( map_matrix, 0, 0, cos(th) ); // row, col, value
  cvmSet( map_matrix, 0, 1, -sin(th) ); // row, col, value
  cvmSet( map_matrix, 0, 2, 0.0 ); // row, col, value
  cvmSet( map_matrix, 1, 0, sin(th) ); // row, col, value
  cvmSet( map_matrix, 1, 1, cos(th) ); // row, col, value
  cvmSet( map_matrix, 1, 2, 0.0 ); // row, col, value
  cvmSet( map_matrix, 2, 0, 0.0 ); // row, col, value
  cvmSet( map_matrix, 2, 1, 0.0 ); // row, col, value
  cvmSet( map_matrix, 2, 2, 1.0 ); // row, col, value
  
  /*
   * MAIN LOOP
   */
  while (!timeToQuitNow)
  {
    /*
     * INPUT IMAGE SETUP
     *
     * the goal here is to populate the _image_ variable with pixels
     *   depending on the source in captureType
     */
    if (captureStatus != STILL) 
    {
      if (captureType == IMAGES_FROM_FOLDER)
      {
        int fileNameMaxSize = 255;
        char filename[fileNameMaxSize];
        
        if (nextFileNumber > MAX_FILE_NUMBER) nextFileNumber = 0;
        sprintf(filename,"%s/image%05d.bmp", FOLDER_NAME, nextFileNumber++);
        inputImage = cvLoadImage(filename, -1); // -1 means determine number of channels from format
      }
      else // captureType == IMAGES_FROM_AVI_FILE or IMAGES_FROM_CAMERA
      {
        frame = cvQueryFrame( capture );  
        if ( !frame  )
        {
          printf( "Frame not obtained when capturing!\n");
          printf( "Setting captureStatus to STILL.\n");
          captureStatus = STILL;
        }
        else  // set up the destination image, if needed
        {
          CvSize frameSize = cvGetSize(frame);
          if ( frameSize.width != inputImage->width || frameSize.height != inputImage->height )
          {
            cvReleaseImage( &inputImage );
            inputImage = cvCreateImage( frameSize, 8, 3 );
            inputImage->origin = frame->origin;
          } // image is OK and the right size, now
          cvCopy( frame, inputImage, 0 );  // copy the frame data into the image
        }
      }
      
      if (captureStatus == ONEFRAME)  captureStatus = STILL; // we got our one frame...
    } 
    
    /*
     * OUTPUT IMAGE SETUP
     *
     * we need to make sure to set up the output images before processing!
     * (or other intermediate images)  we check to be sure they're the right size...
     */
    if ( outputImage == NULL || 
         ( inputImage->width != outputImage->width  || inputImage->height != outputImage->height ) ) 
    { 
      imageAreaInPixels = inputImage->width * inputImage->height; // set this variable...
      // resizes the other images
      if (outputImage) cvReleaseImage( &outputImage );   outputImage = cvCreateImage( cvGetSize(inputImage), 8, 3 );
      if (blu) cvReleaseImage( &blu );  blu = cvCreateImage( cvGetSize(inputImage), 8, 1 ); 
      if (gre) cvReleaseImage( &gre );  gre = cvCreateImage( cvGetSize(inputImage), 8, 1 );
      if (red) cvReleaseImage( &red );  red = cvCreateImage( cvGetSize(inputImage), 8, 1 );
    }
    
    cvCopy(inputImage, outputImage, NULL);  // copy input to output image
    //cvCvtColor( inputImage, hsv, CV_BGR2HSV );  // let's get an hsv image...
    
     
    double avred = 0.0, avgre = 0.0, avblu = 0.0;
    int numpix = 0;
    /*
     * OUR ALGORITHM/PROCESSING/VISUALIZATION
     */
    if (runAlgorithm)
    {
      int b,r,g;
      // loop over the input image
      for ( int x=0 ; x<inputImage->width ; ++x) 
      {
        for ( int y=0 ; y<inputImage->height ; ++y) 
        {
          getrgb( inputImage, x, y, r, g, b );
          if ( currentColor.pixelMatch( r, g, b ) ) 
          {
            avred += r; avgre += g; avblu += b; numpix++;
            setrgb( outputImage, x, y, 0, 255, 0 );
          }
        }
      } // end of for x=0 ...
      
      avred /= numpix; avgre/= numpix; avblu /= numpix;
      printf(" argb: %f %f %f  ", avred, avgre, avblu );
      
    
    
      // split the input images
      cvSplit( inputImage, blu, gre, red, NULL ); // NULL is the alpha channel
      cvInRangeS( inputImage, cvScalar( currentColor.bmin, currentColor.gmin, currentColor.rmin ),  // min side
                              cvScalar( currentColor.bmax, currentColor.gmax, currentColor.rmax ),  // max side
                  red );
      //cvInRangeS( red, cvScalar( currentColor.rmin ), cvScalar( currentColor.rmax ), red );
      cvShowImage( "red", red );
      cvShowImage( "gre", gre );
      cvShowImage( "blu", blu );
      
      // OK, now we blobify!
      blobs = CBlobResult( red, NULL, 100, true ); // source, mask, threshold, find moments?
      // create a file with some of the extracted features
      //blobs.PrintBlobs( "./blobs.txt" );
      //printf("Got %d blobs.\n", blobs.GetNumBlobs() );
      
      // get the top 5 or up to blobs.GetNumBlobs();
      // exclude the top 2
      
      blobs.Filter( blobs, B_INCLUDE, CBlobGetMean(), B_GREATER, 200 );
      blobs.Filter( blobs, B_INCLUDE, CBlobGetArea(), B_GREATER, 100 );
      blobs.Filter( blobs, B_EXCLUDE, CBlobGetDiffY(), B_GREATER, inputImage->height * 0.8 );
      //nthblob = (nthblob + blobs.GetNumBlobs())%blobs.GetNumBlobs();
      //printf("Showing blob #%d.\n", nthblob);
      //blobs.GetNthBlob( CBlobGetHullArea(), nthblob , blobToShow );
      //blobToShow.FillBlob( outputImage, CV_RGB( 255, 255, 0 ));

      double x[5] = { -100.0, -100.0, -100.0 }; // to check if this has already been handled...
      int numSymbols = 0;
      double av = 0.0;
      CBlobGetHullArea getHullPerimeter = CBlobGetHullArea();
      for (int i=0 ; i<MIN(3,blobs.GetNumBlobs()) ; ++i)
      {
        blobs.GetNthBlob( CBlobGetHullArea(), i , blobToShow );
        
        double center = blobToShow.sumx; ///blobToShow.area;
        //printf("#%d c:%f  ", i, center);
        if ( (fabs(center-x[0]) < 30.0) || (fabs(center-x[1]) < 30.0) ) break;
        blobToShow.FillBlob( outputImage, CV_RGB( 255/(i+1), 255/(i+1), 0 ));
        numSymbols++;
        av += getHullPerimeter(blobToShow); //blobToShow.externPerimeter;
        //printf("p: %f ", av);
        x[i] = center;
      }
      av /= numSymbols;
      printf("Saw %d ", numSymbols);
      //printf(" av %f \n", av );
      if (avblu > 15.0) printf(" BLUE ");
      else if (avred > 17.0) printf(" RED ");
      else if (avgre > 20.0) printf(" GREEN ");
      else  printf(" BLUE ");
      if (av > 8000.0) printf(" ELLIPSE ");
      else if (av > 6500.0) printf(" SQUIGGLE ");
      else  printf(" DIAMOND ");
      
      printf("\n");
    
    }
    
    //cvPutText( outputImage, "Hi there!", cvPoint( 20 , 100 ), &ourFont, CV_RGB( 255, 255, 0 ) );
    
    /*
    void cvWarpPerspective( const CvArr* src, CvArr* dst, const CvMat* map_matrix,
                            int flags=CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS,
                            CvScalar fillval=cvScalarAll(0) );
     
     */
    cvWarpPerspective(inputImage,outputImage,map_matrix,CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS,cvScalarAll(0));
     
    
    
    
    /*
		 * DRAW ON THE OUTPUT IMAGE
		 */
		if (drawingRectangle)  // shows a yellow rectangle when selecting a region
		{
		  cvRectangle(outputImage, cvPoint(selection.x,selection.y), 
                  cvPoint(selection.x+selection.width, selection.y+selection.height),
                  cvScalar(0,255,255,0),1,8,0);
		}
    
    cvShowImage( "InputImage", inputImage );    // render the input image
    cvShowImage( "OutputImage", outputImage );  // render the output image
    
    /*
     * HANDLE KEY PRESSES to an OpenCV window
     */
    char cchar = (char) cvWaitKey(10);   // waits for 10 ms to get a keypress
                                         // use cvWaitKey(0) in order to wait forever...
    if ( cchar == 27 )  break;           // break on ESCape key
    
    switch( cchar )                      // handle other key presses
    {        
      case ' ':
      {
        captureStatus = ONEFRAME;
        printf("Getting one frame.\n");
        break;
      }
        
      case 'N':
      {
        nthblob++;
        break;
      }
        
      case 'n':
      {
        nthblob--;
        break;
      }
        
      /*
      case 'S': // save images
      {
        cout << "Save images here... " << endl;
        // practice writing a file with 'Z'
        char filename2[50];
        sprintf(filename2,"./testfile.bmp");
        printf("Wrote file %s\n", filename2);
        cvSaveImage(filename2,outputImage);
        break;
        break;
      }*/
        
      case 'L': // restart live capturing
      {
        captureStatus = LIVE;
        printf("Getting one frame.\n");
        break;
      }
        
      case 'R':
      {
        currentColor.reset();
      }
        
      case 'O': // save params to File
      {
        FILE* fp= fopen("./colordef.txt", "w");
        ColorDefinition* c = &currentColor;
        fprintf(fp, "%s %d %d\n", "r: ", c->rmin, c->rmax);
        fprintf(fp, "%s %d %d\n", "g: ", c->gmin, c->gmax);
        fprintf(fp, "%s %d %d\n", "b: ", c->bmin, c->bmax);
        fclose(fp);
        break;
      }
        
      case 'I':  // get params from file
      {
        FILE* fp=fopen("./colordef.txt", "r");
        ColorDefinition* c = &currentColor;
        fscanf(fp, "r: %d %d\n", &c->rmin, &c->rmax);
        fscanf(fp, "g: %d %d\n", &c->gmin, &c->gmax);
        fscanf(fp, "b: %d %d\n", &c->bmin, &c->bmax);
        currentColor.print();
        fclose(fp);
        break;
      }
        
      case 'A': // turn on or off algorithm running
      {
        if (runAlgorithm) runAlgorithm = false;
        else runAlgorithm = true;
        break;
      }
        
      default:
        break;
    } // end of the keypress-handling switch statement
    
  } // end of the main image input - process - output loop
  
  /*
   * If we're here, some event triggered the main loop to quit
   */
  timeToQuitNow = 1;    // set our signal for the serverThread to quit
  
  printf("Main thread quitting...\n");
  sleep(1.0);             // give it a moment...
  
  /*
   * release the OpenCV resources
   */
  cvReleaseCapture(&capture);
  cvDestroyWindow( "InputImage" );
  cvDestroyWindow( "OutputImage" );
  
  return 0;
}





